==> fxt-2004.10.25.tbz: include/fxt/sort/uniqueapprox.h <==
#if !defined  HAVE_UNIQUEAPPROX_H__
#define       HAVE_UNIQUEAPPROX_H__

#include "fxttypes.h"

template  <typename Type>
inline bool approx_equal(Type x1, Type x2, Type da)
// Return whether abs(x2-x1) <= da
// Must have da>=0
{
    Type d = x2 - x1;
    if ( d<=0 )  d = -d;
    if ( d < da )  return true;
    else           return false;
}
// -------------------------

template  <typename Type>
ulong test_unique_approx(const Type *f, ulong n, Type da)
// For a sorted array test whether all values are
// unique within some tolerance
//   (i.e. whether no value is repeated)
//
// Return 0 if all values are unique
// else return index of the second element in the first pair found
//
// Makes mostly sense with inexact types (float or double)
{
    if ( da<=0 )  da = -da;  // want positive tolerance

    for (ulong k=1; k<n; ++k)
    {
        if ( approx_equal(f[k], f[k-1], da) )  return k;  // k != 0
    }

    return  0;
}
// -------------------------

template  <typename Type>
ulong is_unique_approx(const Type *f, ulong n, Type da)
// For a sorted array test whether all values are unique
// within some tolerance.
//
// Return true if all values are unique, else return false.
{
    return ( 0==test_unique_approx(f, n, da) );
}
// -------------------------


template  <typename Type>
ulong unique_approx_count(const Type *f, ulong n, Type da)
// for a sorted array return the number of unique values
// the number of (not necessarily distinct) repeated
//   values is n - unique_approx_count(f, n, da);
{
    if ( 1>=n )  return n;
    if ( da<=0 )  da = -da;  // Must have positive tolerance
    ulong ct = 1;
    for (ulong k=1; k<n; ++k)
    {
        if ( approx_equal(f[k], f[k-1], da) )  ++ct;
    }
    return  ct;
}
// -------------------------


template  <typename Type>
ulong unique_approx(Type *f, ulong n, Type da)
// For a sorted array squeeze all repeated (within tolerance da) values
// and return the number of unique values.
// Example:  [1, 3, 3, 4, 5, 8, 8] --> [1, 3, 4, 5, 8]
// The routine also works for unsorted arrays as long
// as identical elements only appear in contiguous blocks.
// Example: [4, 4, 3, 7, 7] --> [4, 3, 7]
// The order is preserved.
//.
// Note that the inner while-loop does never access an element out of
// bounds as it is executed only as long as there is at least one
// remaining change of value inside the array.
{
    ulong u = unique_approx_count(f, n, da);

    if ( u==n )  return n;  // nothing to do

    if ( da<=0 )  da = -da;  // Must have positive tolerance
    Type v = f[0];
    for (ulong j=1, k=1;  j<u;  ++j)
    {
        // search next different element:
        while ( approx_equal(f[k], v, da) )
        {
            v = f[k];  // avoid problem with slowly drifting values
            ++k;
        }
        v = f[j] = f[k];
    }

    return u;
}
// -------------------------


#endif  // !defined HAVE_UNIQUEAPPROX_H__

==> fxt-2008.08.17.tbz: include/fxt/sort/uniqueapprox.h <==
#if !defined  HAVE_UNIQUEAPPROX_H__
#define       HAVE_UNIQUEAPPROX_H__

#include "fxttypes.h"

template  <typename Type>
inline bool approx_equal(Type x1, Type x2, Type da)
// Return whether abs(x2-x1) <= da
// Must have da>=0
{
    Type d = x2 - x1;
    if ( d<=0 )  d = -d;
    if ( d < da )  return true;
    else           return false;
}
// -------------------------

template  <typename Type>
ulong test_unique_approx(const Type *f, ulong n, Type da)
// For a sorted array test whether all values are
// unique within some tolerance (i.e. whether no value is repeated).
// Return 0 if all values are unique,
// else return index of the second element in the first pair found.
// Makes mostly sense with inexact types (float or double)
{
    if ( da<=0 )  da = -da;  // want positive tolerance

    for (ulong k=1; k<n; ++k)
    {
        if ( approx_equal(f[k], f[k-1], da) )  return k;  // k != 0
    }

    return  0;
}
// -------------------------

template  <typename Type>
ulong is_unique_approx(const Type *f, ulong n, Type da)
// For a sorted array test whether all values are unique
// within some tolerance.
//
// Return true if all values are unique, else return false.
{
    return ( 0==test_unique_approx(f, n, da) );
}
// -------------------------


template  <typename Type>
ulong unique_approx_count(const Type *f, ulong n, Type da)
// For a sorted array return the number of unique values
// the number of (not necessarily distinct) repeated
//   values is n - unique_approx_count(f, n, da);
{
    if ( 1>=n )  return n;
    if ( da<=0 )  da = -da;  // Must have positive tolerance
    ulong ct = 1;
    for (ulong k=1; k<n; ++k)
    {
        if ( approx_equal(f[k], f[k-1], da) )  ++ct;
    }
    return  ct;
}
// -------------------------


template  <typename Type>
ulong unique_approx(Type *f, ulong n, Type da)
// For a sorted array squeeze all repeated (within tolerance da) values
// and return the number of unique values.
// Example:  [1, 3, 3, 4, 5, 8, 8] --> [1, 3, 4, 5, 8]
// The routine also works for unsorted arrays as long
// as identical elements only appear in contiguous blocks.
// Example: [4, 4, 3, 7, 7] --> [4, 3, 7]
// The order is preserved.
//.
// Note that the inner while-loop does never access an element out of
// bounds as it is executed only as long as there is at least one
// remaining change of value inside the array.
{
    ulong u = unique_approx_count(f, n, da);

    if ( u==n )  return n;  // nothing to do

    if ( da<=0 )  da = -da;  // Must have positive tolerance
    Type v = f[0];
    for (ulong j=1, k=1;  j<u;  ++j)
    {
        // search next different element:
        while ( approx_equal(f[k], v, da) )
        {
            v = f[k];  // avoid problem with slowly drifting values
            ++k;
        }
        v = f[j] = f[k];
    }

    return u;
}
// -------------------------


#endif  // !defined HAVE_UNIQUEAPPROX_H__
